/*------------------------------ array.cc ----------------------------
  The LOOKOUT library encapsulates C/C++ arrays in a family of
  classes which support run-time bounds-checking.
  The classes are IntArray (array of int), FloatArray, DoubleArray,
	and StringArray (array of Lstring).
  The interfaces to all these classes are the same:
     A constructor with two arguments (the second has a default):
	IntArray ia(10,7);  //create an array of 10 ints, each a 7
	FloatArray fa(100); //create an array of 100 floats, each 0.0
     A copy constructor to copy from another array of same type:
     	FloatArray fa2(fa);
     Functions to access the size or length of the array, e.g.:
	fa.size() or ia.length()	
     Overloaded subscript operators which perform bounds checking:
     	ia[9] is OK but ia[10] generates a run-time bounds error
     A sub() function with tells where in a SOURCE program
        a program went out of bounds:
        ia.sub(10) displays the file and line number of ia.sub(10)
        The sub() function is actually a macro which calls _sub().
  Programmer: Glenn D. Blank  
  For chapter 11 of "The Universal Machine"
  Copyright (C) WCB/McGraw-Hill, 1998
---------------------------------------------------------------------*/
#include "array.h"

IntArray::IntArray(int size,int init)
{ //construct size elements with value init
  sz = size;
  element = new int[sz];
  for (int i=0; i < sz; i++)
     element[i] = init;
}

IntArray::IntArray(IntArray & other)
{ //copy constructor copies size and elements from other array
  sz = other.size();
  element = new int[sz];
  for (int i=0; i < sz-1; i++)
    element[i] = other[i];
}

IntArray::~IntArray()
{
  delete [] element;
}

int& IntArray::_sub(char* file, int line, int i) const
{
  if (i < 0 || i >= sz)
  { cout << "\nArray access out of bounds in "
	 << file << ", line " << line
         << endl << "Subscript is " << i  
         << ", but should be in range 0.." << sz-1 << endl;
    exit(1);
  }
  return element[i];
}

int& IntArray::operator[](int i)
{ //overloaded subscript operator returns int element
  if (i < 0 || i >= sz) 
  { cout << "\noperator[" << i << "] is out of bounds!\n"
      << "To locate the error, use sub() member function instead of []\n"
      << "Press Shift-F4 in LOOKOUT to generate a program using sub() function"
      << endl; 
    exit(1);
  }
  return element[i];
}

const int& IntArray::operator[](int i) const
{ //const version of subscript operator avoids side effects
  if (i < 0 || i >= sz)
  { cout << "\noperator[" << i << "] is out of bounds!\n"
      << "To locate the error, use sub() member function instead of []\n"
      << "Press Shift-F4 in LOOKOUT to generate a program using sub() function"
      << endl; 
    exit(1);
  }
  return element[i];
}

int IntArray::size() const
{ return sz; }

int IntArray::length() const
{ return sz; }

FloatArray::FloatArray(int size,float init)
{ //construct size elements with value init
  sz = size;
  element = new float[sz];
  for (int i=0; i < sz; i++)
     element[i] = init;
}

FloatArray::FloatArray(FloatArray & other)
{ //copy constructor copies size and elements from other array
  sz = other.size();
  element = new float[sz];
  for (int i=0; i < sz-1; i++)
    element[i] = other[i];
}

FloatArray::~FloatArray()
{
  delete [] element;
}

float& FloatArray::_sub(char* file, int line, int i) const
{ if (i < 0 || i >= sz) {
    cout << "\nArray access out of bounds in "
	 << file << ", line " << line
         << endl << "Subscript is " << i  
         << ", but should be in range 0.." << sz-1 << endl;
    exit(1);
  }
  return element[i];
}

float& FloatArray::operator[](int i)
{
  if (i < 0 || i >= sz)
  { cout << "\noperator[" << i << "] is out of bounds!\n"
      << "To locate the error, use sub() member function instead of []\n"
      << "Press Shift-F4 in LOOKOUT to generate a program using sub() function"
      << endl; 
    exit(1);
  }
  return element[i];
}

const float& FloatArray::operator[](int i) const 
{
  if (i < 0 || i >= sz)
  { cout << "\noperator[" << i << "] is out of bounds!\n"
      << "To locate the error, use sub() member function instead of []\n"
      << "Press Shift-F4 in LOOKOUT to generate a program using sub() function"
      << endl; 
    exit(1);
  }
  return element[i];
}

int FloatArray::size() const
{ return sz; }

int FloatArray::length() const
{ return sz; }


DoubleArray::DoubleArray(int size,double init)
{ //construct size elements with value init
  sz = size;
  element = new double[sz];
  for (int i=0; i < sz; i++)
     element[i] = init;
}

DoubleArray::DoubleArray(DoubleArray & other)
{ //copy constructor copies size and elements from other array
  sz = other.size();
  element = new double[sz];
  for (int i=0; i < sz-1; i++)
    element[i] = other[i];
}

DoubleArray::~DoubleArray()
{
  delete [] element;
}

double& DoubleArray::_sub(char* file, int line, int i) const
{
  if (i < 0 || i >= sz) 
  { cout << "\nArray access out of bounds in "
	 << file << ", line " << line
         << endl << "Subscript is " << i  
         << ", but should be in range 0.." << sz-1 << endl;
    exit(1);
  }
  return element[i];
}

double& DoubleArray::operator[](int i)
{ 
  if (i < 0 || i >= sz) 
  { cout << "\noperator[" << i << "] is out of bounds!\n"
      << "To locate the error, use sub() member function instead of []\n"
      << "Press Shift-F4 in LOOKOUT to generate a program using sub() function"
      << endl; 
    exit(1);
  }
  return element[i];
}


const double& DoubleArray::operator[](int i) const
{ 
  if (i < 0 || i >= sz)
  { cout << "\noperator[" << i << "] is out of bounds!\n"
      << "To locate the error, use sub() member function instead of []\n"
      << "Press Shift-F4 in LOOKOUT to generate a program using sub() function"
      << endl; 
    exit(1);
  }
  return element[i];
}


int DoubleArray::size() const
{ return sz; }

int DoubleArray::length() const
{ return sz; }


StringArray::StringArray(int size,Lstring init)
{ //construct size elements with value init
  sz = size;
  element = new Lstring[sz];
  for (int i=0; i < sz; i++)
     element[i] = init;
}

StringArray::StringArray(StringArray & other)
{ //copy constructor copies size and elements from other array
  sz = other.size();
  element = new Lstring[sz];
  for (int i=0; i < sz-1; i++)
    element[i] = other[i];
}

StringArray::~StringArray()
{
  delete [] element;
}

Lstring& StringArray::_sub(char* file, int line, int i) const
{
  if (i < 0 || i >= sz)
  { cout << "\nArray access out of bounds in "
	 << file << ", line " << line
         << endl << "Subscript is " << i  
         << ", but should be in range 0.." << sz-1 << endl;
    exit(1);
  }
  return element[i];
}

Lstring& StringArray::operator[](int i)
{
  if (i < 0 || i >= sz)
  { cout << "\noperator[" << i << "] is out of bounds!\n"
      << "To locate the error, use sub() member function instead of []\n"
      << "Press Shift-F4 in LOOKOUT to generate a program using sub() function"
      << endl; 
    exit(1);
  }
  return element[i];
}

const Lstring& StringArray::operator[](int i) const
{
  if (i < 0 || i >= sz)
  { cout << "\noperator[" << i << "] is out of bounds!\n"
      << "To locate the error, use sub() member function instead of []\n"
      << "Press Shift-F4 in LOOKOUT to generate a program using sub() function"
      << endl; 
    exit(1);
  }
  return element[i];
}

int StringArray::size() const
{ return sz; }

int StringArray::length() const
{ return sz; }































